comment %
		    ķ
		    	V86/Enhanced V86 Mode Monitor	
		    Ľ
			  Ŀ
			  	 By Jim Brooks	  
			     Copyright (c) 1995  
			  
%

;------------------------------------------------------------------------------
;   V86/EV86 Mode Monitor 1.0
;
;   Author:	Jim Brooks
;   Date:	21-Nov-1995
;   Internet:	103151.1307@compuserve.com
;
;   You have my permission to copy and distribute this software for
;   non-commercial purposes only. Any commercial use of this software or
;   source code without the author's consent is prohibited.
;
;------------------------------------------------------------------------------

;------------------------------------------------------------------------------
;Requires these LINK options:
;
; /NOF no far call optimization
;
;Warnings:
;
; The V86 exception handlers and their subroutines use ESP to reference the
; the V86 CPL0 stack frame, ESP is expected to point to the V86 EIP.
; Due to these assumptions of stack order, avoid nesting procedures.
;
;Notes:
;
; This is a public-domain derivative of the V86/EV86 component of a
; propietary operating-system. Many features have been removed,
; but vestiges, such as a skeletal task scheduler, still remain.
;
;General register usage:
;
; FS  = FLAT_SEL
; GS  = linear base of task address space
; EDI = task TD
; ESI = flattened EIP (used with GS: for V86 tasks)
; EBP = flattened ESP (used with GS: for V86 tasks)
;
;------------------------------------------------------------------------------

TITLE V86/EV86 Monitor

_VME			= 0				;assemble VME code
_Virtual		= 1				;maintain virtual IF

_Init_V86_EFLAGS	= VMfl or IOPL0 		;IF=0
_Protected_V86_EFLAGS	= 11111111110110110011001000000000b
_Virtual_IF		= 0
_ShootDownUFOs		= 1				;if 1 then kill V86 task if UFO sighted

    IF _Virtual AND (_VME EQ 0)
_Virtual_NonVME         textequ <>
    ELSE
_Virtual_VME		textequ <>
    ENDIF

.586P							;use 486 timings in list

option dotname, noscoped, language:syscall
.seq
.listmacroall

_Emit			= 0				;if false the macro emit does not expand.
_FullList		= 1				;enable full listing file

_PM	textequ <>
include macros.inc

include v86mon.mac
include v86mon.inc

;

V86MonCodeSeg16 segment para public use16 'CODE'
                assume  DS:V86MonDataSeg, SS:V86MonStackSeg, ES:nothing

_CodeSegSize	= 16					;used by emit macros

Entry_V86Mon    label   far

                call    Startup

EnterProtectMode:

		cli					;disable interrupts!
		cld

		mov	ax, V86MonDataSeg
		mov	DS, ax
		mov	ES, ax

		lidt	[IDT_V86_fword]
                lgdt    [GDT_fword]

;Enable protect-mode.

                mov     eax, CR0
                or      eax, pe
                mov     CR0, eax

;Flush prefetch Q since the following instructions were fetched in real-mode.

                jmp     short Flush_Q_of_RM
Flush_Q_of_RM:

;Load TR so that the CPU can identify an outgoing task.

                mov     ax, Init_TSS_SEL
                ltr     ax

;Taskswitch to scheduler.

                jmp     [Task_Scheduler_TaskSwitch]

;

V86MonCodeSeg32 segment para readonly public use32 'CODE'

;			  Ŀ
;			   VME Initialization 
;			  

Entry_Scheduler:

    IF _VME

;Enable VME.

		mov	eax, CR4
		or	eax, VME
		mov	CR4, eax

    ENDIF

;			  Ŀ
;                            Task Scheduler   
;			  

Scheduler:

;Schedule V86 task.

		mov	edi, [TDB]

Scheduler_RunTask:
Scheduler_RunV86Task:

;------------------------------------------------------------------------------
;Expects:
;DS:EDI = V86 TD
;------------------------------------------------------------------------------

;Set SS0:ESP0 of the scheduler's TSS.

		mov	eax, @32 [edi +tss_ss0]
		mov	edx, @32 [edi +tss_esp0]
		mov	@32 [Scheduler_TSS +tss_ss0] , eax
		mov	@32 [Scheduler_TSS +tss_esp0], edx

;Set I/O Map Base Address of the scheduler's TSS to the V86 I/O Bitmap.

		mov	eax, @32 [edi +tss_iobitmapbase_t]
		mov	@32 [Scheduler_TSS +tss_iobitmapbase_t], eax

;Create V86 CPL0 stack frame.

		push	@32 [edi +tss_gs]		;GS
		push	@32 [edi +tss_fs]		;FS
		push	@32 [edi +tss_ds]		;DS
		push	@32 [edi +tss_es]		;ES
		push	@32 [edi +tss_ss]		;SS
		push	@32 [edi +tss_esp]		;ESP
		push	@32 [edi +tss_eflags]		;EFLAGS
		push	@32 [edi +tss_cs]		;CS
		push	@32 [edi +tss_eip]		;EIP

		rdtsc
		mov	[TSC_Start]  , eax
		mov	[TSC_Start+4], edx

;Reload other registers and run V86 task.

		jmp	RunV86Task

;===============

Scheduler_EndV86Task:

		jmp	Scheduler_ExitV86Mon

;===============

Scheduler_ExitV86Mon_UnexpectedError:

		mov	[UnexpectedError], -1
		jmp	Scheduler_ExitV86Mon

;
;
;			  Ŀ
;                            Enter real-mode  
;                         

Scheduler_ExitV86Mon:

EnterRealMode:

;Load segment descriptor cache registers with real-mode counterparts,
;else the protect-mode attributes will remain in effect.

                mov     eax, Real_SEL
                mov     DS, eax
                mov     ES, eax
                mov     FS, eax
                mov     GS, eax
                mov     SS, eax

;The CS segment descriptor cache register is loaded indirectly via a far jump.

                jmp     CS:[ChangeCStoReal]

;---------------
ChangeCStoReal label fword
                dw      AlmostInReal
                dw      0
                dw      Real_Code_SEL
;---------------

;
;
;                     Ŀ
;                         V86 Exception Handlers    
;                     

;-------------------------------------------------------------------------------
comment %
	     31 	       0		    31		      0
	     Ŀ		    Ŀ
TSS.ESP0 --> 		         9*4  TSS.ESP0 --> 		       10*4
	     Ĵ		    Ĵ
	     	       V86 GS   8*4		    	      V86 GS   9*4
	     Ĵ		    Ĵ
	     	       V86 FS   7*4		    	      V86 FS   8*4
	     Ĵ		    Ĵ
	     	       V86 DS   6*4		    	      V86 DS   7*4
	     Ĵ		    Ĵ
	     	       V86 ES   5*4		    	      V86 ES   6*4
	     Ĵ		    Ĵ
	     	       V86 SS   4*4		    	      V86 SS   5*4
	     Ĵ		    Ĵ
	     	   V86 ESP       3*4		    	  V86 ESP       4*4
	     Ĵ		    Ĵ
	     	  V86 EFLAGS     2*4		    	 V86 EFLAGS     3*4
	     Ĵ		    Ĵ
	     	       V86 CS   1*4		    	      V86 CS   2*4
	     Ĵ		    Ĵ
 NEW ESP --> 	   V86 EIP       0*4		    	  V86 EIP       1*4
	     		    Ĵ
					NEW ESP --> 	 Error Code     0*4
						    
%
;-------------------------------------------------------------------------------

;===============

include prefix.asm
include privileg.asm
include io.asm

IF _VME

include iftrans.asm

ELSE

include ioplsens.asm

ENDIF

    IFNDEF _Virtual_VME

;include vpic_386.asm
;include ioplsens.asm

    ELSE

;include vpic_vme.asm
;include iftrans.asm

    ENDIF


;===============

Handler_V86_Debug:
Handler_V86_InvalidOpcode:
Handler_V86_DeviceNotAvailable:
Handler_V86_DoubleFault_IRQ0:
Handler_V86_PageFault:

;-------------------------------------------------------------------------------
;Unimplemented in public domain version.
;-------------------------------------------------------------------------------

		jmp	Scheduler_ExitV86Mon_UnexpectedError

;===============

;                     Ŀ
;                             V86 GPF Handler       
;                     

ALIGN 10h

Handler_V86_GPF:

;Compete V86 task state save.

		push	SS
                pop     DS

		push	edi

		mov	edi, [TDB]

                mov     [edi +tss_eax], eax
                mov     [edi +tss_ebx], ebx
                mov     [edi +tss_ecx], ecx
                mov     [edi +tss_edx], edx
                mov     [edi +tss_ebp], ebp
                mov     [edi +tss_esi], esi
                pop     [edi +tss_edi]; edi

		mov	[Vector], 13

		call	DoErrorCode

;Record the V86 EIP before any adjustments for instruction restartability.

		mov	eax, @32 [esp0_v86_eip]
		mov	[Prefixes], 0			;zero 4 prefix vars at once
		mov	@TD_V86.Unadjusted_EIP, eax

;Flatten relative to V86 linear address:
;GS:ESI = V86 CS:IP
;GS:EBP = V86 SS:SP
;Set GS=linear address of V86 task address space.

		movzx	esi, @16 [esp0_v86_cs]		;V86 CS
		movzx	eax, @16 [esp0_v86_eip] 	;V86 IP
		movzx	ebp, @16 [esp0_v86_ss]		;V86 SS
		movzx	ebx, @16 [esp0_v86_esp] 	;V86 SP
		add	esi, esi			;*2
		add	ebp, ebp			;*2
		mov	GS,  @TD_V86.LinearBase_SEL
		lea	esi, [esi*8 + eax]		;CS*16+IP
		lea	ebp, [ebp*8 + ebx]		;SS*16+IP

V86_GPF_Handler_OneByteOpcode:

;Route GPF via first byte of opcode.

		zero	ebx
		mov	bl, @8 GS:[esi]
		jmp	[ebx*2 + V86_GPF_OneByteOpcode_Routes]

V86_GPF_Handler_TwoByteOpcode:

;The only two-byte opcode of concern is Group7 (0F 01 /n), if not Group7,
;then is a UFO sighting.

		cmp	@8 GS:[esi+1], 01
		je	V86_GPF_Handler_Group7
		jmp	V86_GPF_Handler_UFO

;===============

V86_GPF_Handler_UFO:

;Unidentified faulting opcode has been sighted.

IF _ShootDownUFOs

;Kill V86 task upon UFO sighting.

		mov	[UFO], -1
		mov	ax, GS:[esi]
		mov	[UFO_Opcode], ax

		mov	eax, @32 [esp0_v86_eip]
		mov	[UFO_EIP], eax
		mov	eax, @32 [esp0_v86_cs]
		mov	[UFO_CS], eax
		mov	eax, @32 [esp0_v86_eflags]
		mov	[UFO_EFL], eax

		jmp	Handler_V86_KillTask

ELSE

;Defer unhandled exception #13 to V86 task's ISR.

		call	CompleteV86TaskStateSave

		mov	[Vector], 13

		call	DoErrorCode

		jmp	ReflectInterruptToV86Task

ENDIF

Handler_V86_KillTask:

		jmp	Scheduler_ExitV86Mon

;
;
;		      Ŀ
;                        V86 Interrupt Reflection   
;                     

ALIGN 10h

Handler_V86_Interrupt:

;------------------------------------------------------------------------------
;In order to identify the vector number, each corresponding 32-bit interrupt
;gate has a unique code offset, which routes the interrupts to a table which
;pushes the vector number, and jumps to here.
;------------------------------------------------------------------------------

		call	CompleteV86TaskStateSave

		pop	[Vector]

		;-- fall thru --

ReflectInterruptToV86Task:

;------------------------------------------------------------------------------
;Expects:
;
; SS:ESP points to V86 task's EIP on the CPL0 stack, i.e., any error code
; is expected to have been popped earlier.
;
; [Vector] holds the identified vector number.
;
; V86 CPL0 stack frame.
;------------------------------------------------------------------------------
;Emulate 8086 interrupt semantics:
;
; 1. Push {FLAGS,CS,IP} onto the V86 stack.
; 2. Clear IF,TF in V86 task's EFLAGS.
; 3. Load V86 task's CS,EIP with corresponding 8086 far pointer in the 8086
;    interrupt vector table.
;------------------------------------------------------------------------------

;---------------

;Flatten V86 stack as it would be after pushing {FLAGS,CS,IP}:
;GS:EBP = V86 SS:SP
;Set GS=linear address of V86 task address space.

		movzx	ebp, @16 [esp0_v86_ss]		;V86 SS
		movzx	ebx, @16 [esp0_v86_esp] 	;V86 SP
		add	ebp, ebp			;*2
		mov	GS,  @TD_V86.LinearBase_SEL
		lea	ebp, [ebp*8 + ebx -(2+2+2)]	;SS*16+SP-(2+2+2)

ReflectInterruptToV86Task_SkipFlattenV86Stack:

;------------------------------------------------------------------------------
;Expects:
;
; Flattened V86 SP ( GS:EBP ) has been subtracted by (2+2+2) in preparation
; for pushing {FLAGS,CS,IP}.
;
;This point is jumped to from a GPF subprocedure to bypass reflattening the
;V86 stack.
;------------------------------------------------------------------------------

;Adjust V86 stack in order to push {FLAGS,CS,IP}.

		sub	@32 [esp0_v86_esp], 2+2+2	;V86 ESP

;Emulate pushing {FLAGS,CS,IP} onto the V86 task's stack.

		movzx	eax, @16 [esp0_v86_eip] 	;V86 IP
		movzx	ecx, @16 [esp0_v86_cs]		;V86 CS
		mov	edx, @32 [esp0_v86_eflags]	;V86 FLAGS

IFDEF _Virtual_NonVME

;Substitute IF with @TD_V86.IF for the push FLAGS emulation.

		and	edx, not ifl			;push virtual IF
		or	edx, @TD_V86._IF
ENDIF
IFDEF _Virtual_VME

;Emulate VME semantics of redirecting EFLAGS.VIF to FLAGS_StackImage.IF.

		mov	ebx, edx
		and	ebx, VIF
		shr	ebx, 10 			;VIF -> IF
		and	edx, not ifl
		or	edx, ebx

ENDIF
		mov	@16 GS:[ebp + 0*2], ax		;V86 stack top
		mov	@16 GS:[ebp + 1*2], cx		;V86 stack 2nd
		mov	@16 GS:[ebp + 2*2], dx		;V86 stack 3rd

;Emulate loading the V86 task's CS:IP with the 8086 interrupt vector.
 ;Emulate clearing {IF,TF,AC}.

    IFNDEF _Virtual_VME

		mov	ebx, [Vector]
		 mov	 ecx, @32 [esp0_v86_eflags]
		mov	eax, @32 GS:[ebx*4]
		 and	 ecx, not (ifl or tfl or acfl)
		mov	@16 [esp0_v86_eip], ax
		 or	 ecx, @TD_V86.Virtual_IF
		shr	eax, 16
		 mov	 @32 [esp0_v86_eflags], ecx
		mov	@16 [esp0_v86_cs], ax

;Simulate disabling external interrupts by clearing TD_V86._IF.

		and	@TD_V86._IF, not ifl

    ELSE

		mov	ebx, [Vector]
		mov	eax, @32 GS:[ebx*4]
		mov	@16 [esp0_v86_eip], ax
		shr	eax, 16
		 and	 @32 [esp0_v86_eflags], not (VIF or tfl or acfl)
		mov	@16 [esp0_v86_cs], ax
    ENDIF

		;-- fall thru --

RunV86Task:

;------------------------------------------------------------------------------
;Expects:
; DS:EDI = V86 TSS
; V86 CPL0 stack frame
; PDBR = V86 PD
;
;Reload those V86 registers saved by software instructions from the V86 TSS,
;and use IRETD to reload the rest of the registers and run the V86 task.
;------------------------------------------------------------------------------

		mov	eax, [edi +tss_eax]
                mov     ebx, [edi +tss_ebx]
                mov     ecx, [edi +tss_ecx]
                mov     edx, [edi +tss_edx]
                mov     ebp, [edi +tss_ebp]
                mov     esi, [edi +tss_esi]
                mov     edi, [edi +tss_edi]

                iretd

;===============

ALIGN 10h

CompleteV86TaskStateSave proc

;------------------------------------------------------------------------------
;The CPU has saved these registers of the V86 task onto its CPL0 stack:
; {EIP,CS,EFLAGS,ESP,SS,ES,DS,FS,GS}
;
;Complete task state save by writing these registers into the V86 TSS:
; {EAX,EBX,ECX,EDX,EBP,ESI,EDI}
;
;This procedure sets these registers:
;
; DS  =  monitor's data segment.
; EDI =  Task Definition Base (V86 TSS)
;------------------------------------------------------------------------------

		push	SS
                pop     DS

		push	edi

		mov	edi, [TDB]

                mov     [edi +tss_eax], eax
                mov     [edi +tss_ebx], ebx
                mov     [edi +tss_ecx], ecx
                mov     [edi +tss_edx], edx
                mov     [edi +tss_ebp], ebp
                mov     [edi +tss_esi], esi
                pop     [edi +tss_edi]; edi

                ret

CompleteV86TaskStateSave endp

;===============

DoErrorCode	proc

;------------------------------------------------------------------------------
;Expects:
;
; EDI = V86 TSS
;
; To be called by a handler after it has completed saving the V86 task state.
;
; The handler is not nested within other procedures.
;
; Called only for vectors which correspond to an exception which pushes
; an error code.
;
;Action/return:
;
; Pops off present error code into the variable [ErrorCode].
;
; ZF=1 no error code pushed, which identifies an external interrupt.
;
; ZF=0 error code pushed, which identifies an exception.
;
;This procedure determines if an error code has been pushed by the CPU
;by inferring its presence from the difference of TSS:ESP0 and current ESP.
;If there is an error code there will be 10 pushed double-words, else 9.
;------------------------------------------------------------------------------

		pop	ecx				;pop return address

		mov	eax, [edi +tss_esp0]
		sub	eax, esp

;Difference:
;10*4 = error code pushed
; 9*4 = error code not pushed

		cmp	eax, 9*4

		jz	SkipPopErrorCode

		pop	[ErrorCode]

SkipPopErrorCode:

		jmp	ecx

DoErrorCode	endp

;===============

IF _FullList eq 0
.nolist
ENDIF

include vectors.inc

.list

;===============

;


;===============

V86MonCodeSeg32End:

V86MonCodeSeg32 ends

;Set the data segment limit at the program source end, for consideration
;of the EMITx macros which append to the data segment.

V86MonDataSeg   segment para public 'DATA'
                db ?

V86MonDataSegEnd = $

V86MonDataSeg   ends

;===============

include 8086.asm
include discard.asm

;===============

V86MonCodeSeg16End = $

V86MonCodeSeg16 ends

end Entry_V86Mon
